﻿-- Author      : RisM
-- Create Date : 11/28/2009 11:13:26 PM


local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)

SpeakinSpell:PrintLoading("network.lua")


-------------------------------------------------------------------------------
-- NETWORK VERSION
-------------------------------------------------------------------------------
-- Increment this only when the format changes badly enough to break older clients
-- 3.2.2.21 added a check for this version mismatch
-- all packets MUST pass: packet.commversion = SpeakinSpell.NETWORK_VERSION
-- NOTE: Adding new command codes does NOT require incrementing this version

SpeakinSpell.NETWORK_VERSION = "3.2.2.21" -- CHANGE WITH CARE! SEE ABOVE!

SpeakinSpell.DEBUG_NETWORK = false --controls whether to accept messages from yourself, mostly


-------------------------------------------------------------------------------
-- INVISIBLE NETWORK COMMUNICATIONS BETWEEN INSTANCES OF SPEAKINSPLL
-------------------------------------------------------------------------------
-- TODOFUTURE: add a global channel or friend list network of known SS users


function SpeakinSpell:Network_Init()
	if not SpeakinSpellSavedDataForAll.Networking.EnableNetwork then
		return
	end
		
	self:RegisterComm("SPEAKINSPELL", "Network_OnCommReceive")
	
	-- auto-sync at startup
	if SpeakinSpellSavedDataForAll.Networking.AutoSyncOnLogin then
		self:Network_AutoSync()
	end
end


function SpeakinSpell:Network_Disable()
	self:UnregisterComm("SPEAKINSPELL")
end


-------------------------------------------------------------------------------
-- STATUS REPORTS
-------------------------------------------------------------------------------


function SpeakinSpell:Network_PrintStatus( rxtx, msg, de )
	local funcname = "Network_PrintStatus"
	
	-- Built-in status messages
	-- msg can be nil because some messages are optional below ShowCommTraffic, i.e. ShowTransferProgress
	if msg then
		if SpeakinSpellSavedDataForAll.Networking.ShowCommTraffic then
			self:SayMultiLineWithSubs(msg, "COMM TRAFFIC "..rxtx, nil, false, de)
		else
			self:DebugMsg(funcname, msg)
		end
	end

	-- if the network status event has a name, then signal it as a speech event
	if de.name then
		de.type = "COMM"..rxtx
		de.msg = msg
		self:OnSpeechEvent( de )
	end
end


-------------------------------------------------------------------------------
-- START SYNC
-------------------------------------------------------------------------------


function SpeakinSpell:Network_SyncWithTarget(target)
	local de = {
		name = L["Auto-Sync With Player"],
		target = target,
	}
	self:Network_PrintStatus( "TX", L["Starting Sync with <target>"], de )
	
	local data = {
		Share	= SpeakinSpellSavedDataForAll.Networking.Share,
		Collect	= SpeakinSpellSavedDataForAll.Networking.Collect,
	}
	self:Network_SendCommand("SYNC?", data, "WHISPER", target)
end


function SpeakinSpell:Network_AutoSync() --TODO: call if AutoSyncOnJoin and we join one of the channels
	local de = {
		name = L["General Auto-Sync"],
	}
	self:Network_PrintStatus( "TX", L["Performing General Auto-Sync"], de )

	local data = {
		Share	= SpeakinSpellSavedDataForAll.Networking.Share,
		Collect	= SpeakinSpellSavedDataForAll.Networking.Collect,
	}
	self:Network_BroadcastCommand("SYNC?", data)

	-- automatic debug test suite if enabled
	if SpeakinSpell.DEBUG_NETWORK then
		data.big_extra_data = SpeakinSpellSavedData
		self:Network_SendCommand("SYNC?",data,"GUILD")
		self:Network_SendCommand("SYNC?",data,"RAID")
		self:Network_SendCommand("SYNC?",data,"PARTY")
		self:Network_SendCommand("SYNC?",data,"BATTLEGROUND")
		self:Network_SendCommand("SYNC?",data,"WHISPER", "nobody")
		self:Network_SendCommand("SYNC?",data,"WHISPER", UnitName("player"))
	end
end


-------------------------------------------------------------------------------
-- COMMUNICATION PROTOCOL - COMMAND / RESPONSE HANDLERS
-------------------------------------------------------------------------------

-- route commands to functions
local NETWORK_COMMANDS = {

-------------------------------------------------------------------------------
["SYNC?"] = function( packet )
	local sentsomething = false
	
	if not packet.data then
		-- unexpected missing packet data, just reply with version
		SpeakinSpell:Network_SendReply( "V=", nil, packet )
		return
	end
	
	-- Ask for event tables, if I collect them, and the sender shares them
	if SpeakinSpellSavedDataForAll.Networking.Collect.ET and packet.data.Share.ET then
		SpeakinSpell:Network_SendReply( "ET?", nil, packet )
		sentsomething = true
	end
	
	-- If I share my ET and the sender wants it, send it now
	if SpeakinSpellSavedDataForAll.Networking.Share.ET and packet.data.Collect.ET then
		SpeakinSpell:Network_SendReply( "ET=", SpeakinSpell:GetActiveEventTable(), packet )
		sentsomething = true
	end
	
	-- Ask for event tables, if I collect them, and the sender shares them
	if SpeakinSpellSavedDataForAll.Networking.Collect.RS and packet.data.Share.RS then
		SpeakinSpell:Network_SendReply( "RS?", nil, packet )
		sentsomething = true
	end
	
	-- If I share my RS and the sender wants it, send it now
	if SpeakinSpellSavedDataForAll.Networking.Share.RS and packet.data.Collect.RS then
		SpeakinSpell:Network_SendReply( "RS=", SpeakinSpellSavedDataForAll.RandomSubs, packet )
		sentsomething = true
	end
	
	-- Ask for NewEventsDetected, if I collect them, and the sender shares them
	-- If I share my NewEventsDetected and the sender wants it, send it now
	-- NOTE: if I'm sending my list and asking for a reply, I'll attach my list to the NEW? and get a shorter reply
	--		instead of asking for a separate NEW? and NEW= and getting back a mostly redundant list
	if SpeakinSpellSavedDataForAll.Networking.Collect.NEW and packet.data.Share.NEW then
		if SpeakinSpellSavedDataForAll.Networking.Share.NEW and packet.data.Collect.NEW then
			-- here's my list and I want a reply
			SpeakinSpell:Network_SendReply( "NEW?", SpeakinSpellSavedDataForAll.NewEventsDetected, packet )
		else
			-- I'd like a list from you, but I don't share mine
			SpeakinSpell:Network_SendReply( "NEW?", nil, packet )
		end
		sentsomething = true
	elseif SpeakinSpellSavedDataForAll.Networking.Share.NEW and packet.data.Collect.NEW then
		-- here's my list, and I don't want a reply
		SpeakinSpell:Network_SendReply( "NEW=", SpeakinSpellSavedDataForAll.NewEventsDetected, packet )
		sentsomething = true
	end
	
	-- if I haven't said anything else yet
	-- at least brag that I'm a newer version
	if (not sentsomething) and (SpeakinSpell.CURRENT_VERSION > packet.clientversion) then
		SpeakinSpell:Network_SendReply( "V=", nil, packet )
	end
	
end,

-------------------------------------------------------------------------------
["V="] = function( packet )
	-- this is done for every packet received, including this one - see OnCommReceive
	--	SpeakinSpell:Network_OnCheckNewVersionReport( packet.clientversion )
	-- stubbed here just to have a valid entry for "V="
end,

-------------------------------------------------------------------------------
["ET?"] = function( packet )
	if SpeakinSpellSavedDataForAll.Networking.Share.ET then
		SpeakinSpell:Network_SendReply( "ET=", SpeakinSpell:GetActiveEventTable(), packet )
	end
end,

-------------------------------------------------------------------------------
["ET="] = function( packet )
	SpeakinSpell:Network_OnEventTable( packet )
end,

-------------------------------------------------------------------------------
["RS?"] = function( packet )
	if SpeakinSpellSavedDataForAll.Networking.Share.RS then
		SpeakinSpell:Network_SendReply( "RS=", SpeakinSpellSavedDataForAll.RandomSubs, packet )
	end
end,

-------------------------------------------------------------------------------
["RS="] = function( packet )
	SpeakinSpell:Network_OnRandomSubs( packet )
end,

-------------------------------------------------------------------------------
["NEW?"] = function( packet )
	-- send the NewEventsDetected
	local replydata = SpeakinSpellSavedDataForAll.NewEventsDetected
	
	-- if a table of known events was provided, then don't send those to a client that already knows about them
	if packet.data then
		-- process the received data like it's a "NEW="
		SpeakinSpell:Network_OnNewEvents( packet )
		-- format a reply that is reduced in size to remove events the far end already knows about
		replydata = {}
		for key,de in pairs(SpeakinSpellSavedDataForAll.NewEventsDetected) do
			local known = packet.data[key]
			if not known then
				replydata[key] = de
			end
		end
	end
	
	-- if there are no new events, then don't bother to reply
	if not SpeakinSpell:IsTableEmpty(replydata) then
		SpeakinSpell:Network_SendReply("NEW=", replydata, packet)
	else -- report nothing to send
		local de = {
			name = L["No New Events to Send to <target>"],
			target = packet.sender,
		}
		SpeakinSpell:Network_PrintStatus("TX", L["reply canceled - all event hooks are already known to <target>"], de)
		return
	end
end,

-------------------------------------------------------------------------------
["NEW="] = function( packet )
	SpeakinSpell:Network_OnNewEvents( packet )
end,

-------------------------------------------------------------------------------

} --END NETWORK_COMMANDS
	


-------------------------------------------------------------------------------
-- BASIC SEND FUNCTIONS
-------------------------------------------------------------------------------


function SpeakinSpell:Network_SendCommand(command, data, distribution, target)
	local funcname = "Network_SendCommand"
	
	-- Signal comm traffic event
	local de = {
		name = L["Send Data"],
		target = target or distribution,--target name for whispers, or a channel name
		caster = UnitName("player"),
		commandcode = command,
		distribution = distribution,
	}
	if data then
		de.data = tostring(data)
	else
		de.data = ""
	end
	self:Network_PrintStatus( "TX", L["Send to <target> <commandcode><data>"], de )
	
	-- setup callback function
	local callbackArg = {
		command = command,
		distribution = distribution,
		target = target,
	}
	--TODO: how do you pass this (correctly) directly as a function pointer?
	local callbackFn = function(callbackArg, sent, textlen)
		return self:Network_OnPacketSent(callbackArg, sent, textlen)
	end
	
	-- format the packet
	-- it's easier to maintain and read the code if we serialize it into a single table
	-- even if it does add a few bytes to specify the parameter names
	local packet = {
		-- from me
		sender = UnitName("player"),
		-- version info
		dataversion = SpeakinSpell.DATA_VERSION,
		clientversion = SpeakinSpell.CURRENT_VERSION,
		commversion = SpeakinSpell.NETWORK_VERSION, -- past versions are checking this
		-- command and parameters
		command = command,
		data = data,
	}
	local packetstring = self:Serialize( packet )
	
	-- send
	self:SendCommMessage("SPEAKINSPELL", packetstring, distribution, target, "BULK", callbackFn, callbackArg)
end


function SpeakinSpell:Network_SendReply(command, data, packet)
	if packet.distribution == "WHISPER" then
		self:Network_SendCommand(command, data, packet.distribution, packet.sender)
	else -- reply to the whole channel
		self:Network_SendCommand(command, data, packet.distribution, nil)
	end
end


function SpeakinSpell:Network_OnPacketSent(callbackArg, sent, textlen)
	local funcname = "Network_OnPacketSent"
	
	-- show progress
	local de = {
		name = L["Each Packet Sent"],
		target	= callbackArg.target or callbackArg.distribution,
		command = callbackArg.command,
		distribution= callbackArg.distribution,
		sent = sent,
		textlen = textlen,
		percent = string.format( "%02d", (100*sent/textlen) ),
	}
	local msg = L["<target> -> <command> <sent> / <textlen> - <percent>%"]
	if not SpeakinSpellSavedDataForAll.Networking.ShowTransferProgress then
		msg = nil
	end
	self:Network_PrintStatus( "TX", msg, de )
	
	-- check for transfer complete
	if sent == textlen then
		de.name = L["Transfer Complete"]
		self:Network_PrintStatus( "TX", L["Send Complete: <command> -> <target>"], de )
	end
end


-------------------------------------------------------------------------------
-- Network_OnCommReceive RECEIVE FUNCTION
-------------------------------------------------------------------------------


function SpeakinSpell:Network_OnCommReceive(prefix, Msgs, distribution, target)
	local funcname = "Network_OnCommReceive"
	
	-- all messages are packaged in a single table called packet
	local success, packet = self:Deserialize(Msgs)
	if not (success and type(packet) == "table") then
		local de = {
			name = L["Network Error"],
		}
		self:Network_PrintStatus( "RX", L["Network Error: Deserialize failed"], de )
		return
	end
	
	if packet.commversion ~= SpeakinSpell.NETWORK_VERSION then
		-- incompatible version of the comm system at a critical level
		-- this is not something I plan to change often, but will be necessary if I change it again like I am now
		-- to stop older versions from throwing LUA errors
		-- this should only be done if the packet table format has changed in a way that would throw LUA errors in older code
		-- otherwise, change command code strings to route to newer functions, and allow older clients to show unknown command errors
		
		-- show that a new version is available
		self:Network_CheckForNewVersionReport( packet.commversion )

		-- show network error for debug output
		local de = {
			name = L["Version Mismatch Error"],
			commversion = commversion,
		}
		self:Network_PrintStatus( "RX", L["Network Error: sender has incompatible version <commversion>"], de )
		
		return
	end
	
	-- ignore messages from myself
	if packet.sender == UnitName("player") and not SpeakinSpell.DEBUG_NETWORK then
		self:DebugMsg(funcname, "sender is self")
		return
	end
	
	-- remember the distribution with the packet info as we pass it to a function handler
	packet.distribution = distribution
	
	-- show comm progress message
	local de = {
		name = L["Receive Data"],
		target = UnitName("player"),
		caster = packet.sender,
		sender = packet.sender,
		commandcode = packet.command,
		distribution = distribution,
	}
	if packet.data then
		de.data = tostring(packet.data)
	else
		de.data = ""
	end
	self:Network_PrintStatus( "RX", L["Received from <sender> (<distribution>): <commandcode><data>"], de )

	-- always check for a new version
	-- V= packets are only sent if we can't think of anything else to say
	self:Network_CheckForNewVersionReport( packet.clientversion )
	
	-- find a handler for this command
	local func = NETWORK_COMMANDS[packet.command]
	
	-- bad command code?
	if not func then
		local de = {
			name = L["Invalid Command Code"],
			code = packet.command,
			sender = packet.sender,
			version = packet.clientversion,
		}
		self:Network_PrintStatus( "RX", L["Network Error: received unexpected code \"<code>\" from <sender> v<version>"], de )
		return
	end
	
	-- process the packet through it's command handler in NETWORK_COMMANDS above
	func( packet )
end



-------------------------------------------------------------------------------
-- BROADCASTING AND UTILS
-------------------------------------------------------------------------------



function SpeakinSpell:Network_GetBroadcastChannels()
	-- only use channels that apply to this character
	-- otherwise we could generate "you are not in a guild" spam, and the like
	local BroadcastChannels = {
		--"GUILD",
		--"RAID",
		--"PARTY",
	}
	if GetGuildInfo( UnitName("player") ) then
		tinsert(BroadcastChannels, "GUILD")
	end
	if GetNumPartyMembers() > 0 then
		tinsert(BroadcastChannels, "PARTY")
	end
	if GetNumRaidMembers() > 0 then
		tinsert(BroadcastChannels, "RAID")
	end
	if self:IsInBG() then
		tinsert(BroadcastChannels, "BATTLEGROUND")
	end
	return BroadcastChannels
end


-- send a command to all possible channels
function SpeakinSpell:Network_BroadcastCommand(command, data)
	-- where should we broadcast?
	local BroadcastChannels = self:Network_GetBroadcastChannels()
	
	-- report the failure if no broadcast channels are available
	if self:IsTableEmpty( BroadcastChannels ) then
		local de = {
			name = L["General Auto-Sync Failed"],
		}
		self:Network_PrintStatus("TX", L["No channels are available. For global syncs, join a guild, party, raid, or battleground. Otherwise, sync with a specific target."], de)
		return
	end
	
	-- send to all channels
	for i,channel in pairs(BroadcastChannels) do
		self:Network_SendCommand(command, data, channel, nil)
	end
end


-------------------------------------------------------------------------------
-- VERSION NUMBER EXCHANGE
-------------------------------------------------------------------------------


function SpeakinSpell:Network_CheckForNewVersionReport( version )
	version = tostring(version)
	if version > SpeakinSpell.RuntimeData.Networking.NewestVersion then
		-- an upgrade is available
		SpeakinSpell.RuntimeData.Networking.NewestVersion = version
		self:Print(L["Newer version available: "]..version)
		self:Print( SpeakinSpell.URL )
	end
end


-------------------------------------------------------------------------------
-- NEW EVENTS DETECTED
-------------------------------------------------------------------------------


function SpeakinSpell:Network_ValidateNewEvents( packet )
	local funcname = "Network_ValidateNewEvents"
	
	-- if this came from a client newer than us, then we must disregard it because we have no idea what could be new
	if packet.dataversion > SpeakinSpell.DATA_VERSION then
		local de = {
			name = L["Ignored New Events List data from a newer version"],
			sender = packet.sender,
			dataversion = packet.dataversion,
		}
		-- NOTE: this error occurs in the older client, 
		--		which can not have foreknowledge to know about the reason for a change to the data version in the newer client
		--		while it's possible that the data change was unrelated to the NewEventsDetected table
		--		we're not going to take that risk - just make the user update to the newer client
		--		the data version <> client version separation already allows for enough flexibility between different client versions
		self:Network_PrintStatus("RX", L["I'm afraid to import New Events List data from <sender's> newer client (v<dataversion>+)"], de)
		return false
	end
	
	-- check for expected data types
	-- NewEvents = (table)SpeakinSpellSavedDataForAll.NewEventsDetected,
	-- WantReply = (bool)WantReply
	if type(data) ~= "table" then
		self:DebugMsg(funcname, "type(data):"..type(data))
		return false
	end
	
	-- NOTE: we're not going to get into supporting applying patches to comm with old clients
	--		if they're not using the current data version, they need to upgrade
	--		levereging the ApplyPatches system is theoretically possible, but unnecessary
	
	-- NOTE: don't bother to Validate_DetectedEvent until we detect a new key
	return true
end


function SpeakinSpell:Network_OnNewEvents( packet )
	local funcname = "Network_OnNewEvents"
	
	-- merge new events lists
	-- (replies handled elsewhere)
	
	-- disregard this if we don't collect this data
	if not SpeakinSpellSavedDataForAll.Networking.Collect.NEW then
		self:DebugMsg(funcname, "CollectNEW is disabled")
		return
	end
	
	-- check for expected parameters
	-- and update data version format from older clients if necessary
	if not self:Network_ValidateNewEvents( packet ) then
		return
	end
		
	-- import the new events we just received
	for key,de in pairs(packet.data) do
		local known = SpeakinSpellSavedDataForAll.NewEventsDetected[key]
		if not known then
			self:Validate_DetectedEvent(de)
			if key ~= de.key then
				self:DebugMsg(funcname, "validated de has key mismatch, disregarding it")
			else
				-- add the new event info
				SpeakinSpellSavedDataForAll.NewEventsDetected[key] = de
				
				-- force the GUI to update
				self.RuntimeData.OptionsGUIStates.CreateNew.FilterChanged = true
				
				-- report status
				--self:DebugMsg(funcname, "new key:"..tostring(key))
				local de_for_msg = self:CopyTable(de)
				de_for_msg.name = nil -- don't signal each NewEventsDetected as a speech event - they come in big sets, a bunch at a time
				de_for_msg.sender = packet.sender
				self:Network_PrintStatus( "RX", L["Discovered new event hook: <displayname>"], de_for_msg )

			end
		end
	end
end


-------------------------------------------------------------------------------
-- EVENT TABLES for CONTENT PACKS
-------------------------------------------------------------------------------


function SpeakinSpell:Network_ValidateEventTable( packet )
	local funcname = "Network_ValidateEventTable"
	
	-- if this came from a client newer than us, then we must disregard it because we have no idea what could be new
	if packet.dataversion > SpeakinSpell.DATA_VERSION then
		local de = {
			name = L["Ignored Speech List data from a newer version"],
			sender = packet.sender,
			dataversion = packet.dataversion,
		}
		-- NOTE: this error occurs in the older client, 
		--		which can not have foreknowledge to know about the reason for a change to the data version in the newer client
		--		while it's possible that the data change was unrelated to the NewEventsDetected table
		--		we're not going to take that risk - just make the user update to the newer client
		--		the data version <> client version separation already allows for enough flexibility between different client versions
		self:Network_PrintStatus("RX", L["I'm afraid to import Speeches data from <sender's> newer client (v<dataversion>+)"], de)
		return false
	end
	
	-- check for expected data types
	-- data = EventTable
	if type(packet.data) ~= "table" then
		self:DebugMsg(funcname, "type(data):"..type(data))
		return false
	end
	
	-- NOTE: we're not going to get into supporting applying patches to comm with old clients
	--		if they're not using the current data version, they need to upgrade
	--		levereging the ApplyPatches system is theoretically possible, but unnecessary
	
	-- also run validation functions
	-- new data added via default data structure templates sometimes do not get patch functions
	for key,ete in pairs( packet.data ) do
		self:Validate_EventTableEntry(ete)
	end
	
	return true
end


function SpeakinSpell:Network_OnEventTable( packet )
	local funcname = "Network_OnEventTable"
	
	-- disregard this if we don't collect speeches
	if not SpeakinSpellSavedDataForAll.Networking.Collect.ET then
		self:DebugMsg(funcname, "CollectET disabled")
		return
	end
	
	-- validate data
	if not self:Network_ValidateEventTable( packet ) then
		return
	end
	
	-- save the content for future use by the templates system and /ss import
	SpeakinSpellSavedDataForAll.Networking.CollectedEventTables[packet.sender] = packet.data

	local de = {
		name = L["Collected Event Table (Speeches)"],
		sender = packet.sender,
	}
	self:Network_PrintStatus( "RX", L["Collected an Event Table from: <sender>"], de )
end


function SpeakinSpell:Network_OnRandomSubs( packet )
	-- disregard this if we don't collect RandomSubs
	if not SpeakinSpellSavedDataForAll.Networking.Collect.RS then
		self:DebugMsg(funcname, "CollectRS disabled")
		return
	end
	
	-- save the content for future use by the templates system and /ss import
	SpeakinSpellSavedDataForAll.Networking.CollectedRandomSubs[packet.sender] = packet.data

	local de = {
		name = L["Collected Random Substitutions"],
		sender = packet.sender,
	}
	self:Network_PrintStatus( "RX", L["Collected random substitutions from: <sender>"], de )
end

